#include "MonochromeLib.h"

#include "draw.h"
#include "level.h"
#include "object.h"
#include "tileset.h"

/* global variables */
Tileset *ground, *ground_mask[4], *spike[4], *egg[4], *egg_caught, *duck[4][4], *gravity_switch, *cube, *caterpillar[4][4], *plant[4], *bullet, *bee[4], *bee_halo, *pestle[4];

/* public functions */
int draw_init()
{
	ground = tileset_load(T_GROUND);

	ground_mask[DOWN] = tileset_load(GROUND_MASK_DOWN);
	ground_mask[LEFT] = tileset_load(GROUND_MASK_LEFT);
	ground_mask[RIGHT] = tileset_load(GROUND_MASK_RIGHT);

	spike[UP] = tileset_load(T_SPIKE_UP);
	spike[DOWN] = tileset_load(T_SPIKE_DOWN);
	spike[LEFT] = tileset_load(T_SPIKE_LEFT);
	spike[RIGHT] = tileset_load(T_SPIKE_RIGHT);

	egg[UP] = tileset_load(T_EGG_UP);
	egg[DOWN] = tileset_load(T_EGG_DOWN);
	egg[LEFT] = tileset_load(T_EGG_LEFT);
	egg[RIGHT] = tileset_load(T_EGG_RIGHT);
	egg_caught = tileset_load(EGG_CAUGHT);

	duck[UP][RIGHT] = tileset_load(DUCK_UP_RIGHT);
	duck[UP][LEFT] = tileset_load(DUCK_UP_LEFT);
	duck[DOWN][RIGHT] = tileset_load(DUCK_DOWN_RIGHT);
	duck[DOWN][LEFT] = tileset_load(DUCK_DOWN_LEFT);
	duck[RIGHT][UP] = tileset_load(DUCK_RIGHT_UP);
	duck[RIGHT][DOWN] = tileset_load(DUCK_RIGHT_DOWN);
	duck[LEFT][UP] = tileset_load(DUCK_LEFT_UP);
	duck[LEFT][DOWN] = tileset_load(DUCK_LEFT_DOWN);

	gravity_switch = tileset_load(T_GRAVITY_SWITCH);
	cube = tileset_load(T_CUBE);

	caterpillar[UP][RIGHT] = tileset_load(CATERPILLAR_UP_RIGHT);
	caterpillar[UP][LEFT] = tileset_load(CATERPILLAR_UP_LEFT);
	caterpillar[DOWN][RIGHT] = tileset_load(CATERPILLAR_DOWN_RIGHT);
	caterpillar[DOWN][LEFT] = tileset_load(CATERPILLAR_DOWN_LEFT);
	caterpillar[RIGHT][UP] = tileset_load(CATERPILLAR_RIGHT_UP);
	caterpillar[RIGHT][DOWN] = tileset_load(CATERPILLAR_RIGHT_DOWN);
	caterpillar[LEFT][UP] = tileset_load(CATERPILLAR_LEFT_UP);
	caterpillar[LEFT][DOWN] = tileset_load(CATERPILLAR_LEFT_DOWN);

	plant[LEFT] = tileset_load(T_PLANT_LEFT);
	plant[RIGHT] = tileset_load(T_PLANT_RIGHT);

	bullet = tileset_load(T_BULLET);

	bee[LEFT] = tileset_load(T_BEE_LEFT);
	bee[RIGHT] = tileset_load(T_BEE_RIGHT);
	bee_halo = tileset_load(BEE_HALO);

	pestle[UP] = tileset_load(T_PESTLE_UP);
	pestle[DOWN] = tileset_load(T_PESTLE_DOWN);
	pestle[LEFT] = tileset_load(T_PESTLE_LEFT);
	pestle[RIGHT] = tileset_load(T_PESTLE_RIGHT);

	//verify if all tilesets are correctly loaded
	return 0;
}

void draw_quit()
{
	tileset_free(ground);

	tileset_free(ground_mask[DOWN]);
	tileset_free(ground_mask[LEFT]);
	tileset_free(ground_mask[RIGHT]);

	tileset_free(spike[UP]);
	tileset_free(spike[DOWN]);
	tileset_free(spike[LEFT]);
	tileset_free(spike[RIGHT]);

	tileset_free(egg[UP]);
	tileset_free(egg[DOWN]);
	tileset_free(egg[LEFT]);
	tileset_free(egg[RIGHT]);
	tileset_free(egg_caught);

	tileset_free(duck[UP][RIGHT]);
	tileset_free(duck[UP][LEFT]);
	tileset_free(duck[DOWN][RIGHT]);
	tileset_free(duck[DOWN][LEFT]);
	tileset_free(duck[RIGHT][UP]);
	tileset_free(duck[RIGHT][DOWN]);
	tileset_free(duck[LEFT][UP]);
	tileset_free(duck[LEFT][DOWN]);

	tileset_free(gravity_switch);
	tileset_free(cube);

	tileset_free(caterpillar[UP][RIGHT]);
	tileset_free(caterpillar[UP][LEFT]);
	tileset_free(caterpillar[DOWN][RIGHT]);
	tileset_free(caterpillar[DOWN][LEFT]);
	tileset_free(caterpillar[RIGHT][UP]);
	tileset_free(caterpillar[RIGHT][DOWN]);
	tileset_free(caterpillar[LEFT][UP]);
	tileset_free(caterpillar[LEFT][DOWN]);
	
	tileset_free(plant[LEFT]);
	tileset_free(plant[RIGHT]);

	tileset_free(bullet);

	tileset_free(bee[LEFT]);
	tileset_free(bee[RIGHT]);
	tileset_free(bee_halo);

	tileset_free(pestle[UP]);
	tileset_free(pestle[DOWN]);
	tileset_free(pestle[LEFT]);
	tileset_free(pestle[RIGHT]);
}

void draw_tileMap(Level* level, Camera camera)
{
	int i, j, x, y;
	for(j=0 ; j<SCREEN_HEIGHT/TILE_HEIGHT+1 ; j++) {
		y = TILE_HEIGHT*j - (camera.y&TILE_HEIGHT-1);
		for(i=0 ; i<SCREEN_WIDTH/TILE_WIDTH+1 ; i++) {
			x = TILE_WIDTH*i - (camera.x&TILE_WIDTH-1);
			if((camera.x>>4)+i>=0 && (camera.y>>4)+j>=0 && (camera.x>>4)+i<level->width && (camera.y>>4)+j<level->height) {
				switch(level->map[((camera.y>>4)+j)*level->width + (camera.x>>4)+i]) {
					case GROUND:
						ML_bmp_or_cl(ground->tile[0], x, y, TILE_WIDTH, TILE_HEIGHT);
						if((camera.y>>4)+j > 0)
							if(level->map[((camera.y>>4)+j-1)*level->width + (camera.x>>4)+i] != GROUND)
								ML_bmp_and_cl(ground->tile[1], x, y, TILE_WIDTH, TILE_HEIGHT);
						if((camera.y>>4)+j < level->height-1)
							if(level->map[((camera.y>>4)+j+1)*level->width + (camera.x>>4)+i] != GROUND)
								ML_bmp_and_cl(ground_mask[DOWN]->tile[0], x, y, TILE_WIDTH, TILE_HEIGHT);
						if((camera.x>>4)+i < level->width-1)
							if(level->map[((camera.y>>4)+j)*level->width + (camera.x>>4)+i+1]!= GROUND)
								ML_bmp_and_cl(ground_mask[RIGHT]->tile[0], x, y, TILE_WIDTH, TILE_HEIGHT);
						if((camera.x>>4)+i > 0)
							if(level->map[((camera.y>>4)+j)*level->width + (camera.x>>4)+i-1] != GROUND)
								ML_bmp_and_cl(ground_mask[LEFT]->tile[0], x, y, TILE_WIDTH, TILE_HEIGHT);
						break;
					case SPIKE_UP: ML_bmp_or_cl(spike[UP]->tile[0], x, y, TILE_WIDTH, TILE_HEIGHT); break;
					case SPIKE_DOWN: ML_bmp_or_cl(spike[DOWN]->tile[0], x, y, TILE_WIDTH, TILE_HEIGHT); break;
					case SPIKE_LEFT: ML_bmp_or_cl(spike[LEFT]->tile[0], x, y, TILE_WIDTH, TILE_HEIGHT); break;
					case SPIKE_RIGHT: ML_bmp_or_cl(spike[RIGHT]->tile[0], x, y, TILE_WIDTH, TILE_HEIGHT); break;
				}
			} else ML_bmp_or_cl(ground->tile[0], x, y, TILE_WIDTH, TILE_HEIGHT);
		}
	}
}

void draw_objects(Level* level, Camera camera)
{
	int i;
	Object* object;
	for(i=0 ; level->object[i] ; i++) {
		object = level->object[i];
		switch(object->type) {
			case DUCK: ML_bmp_or_cl(duck[object->head][object->look]->tile[object->position], object->x-camera.x, object->y-camera.y, duck[object->head][object->look]->width, duck[object->head][object->look]->height); break;
			case EGG:
				if(object->position) ML_bmp_or_cl(egg_caught->tile[object->position-1], object->x-camera.x, object->y-camera.y, egg_caught->width, egg_caught->height);
				else ML_bmp_or_cl(egg[object->head]->tile[0], object->x-camera.x, object->y-camera.y, egg[object->head]->width, egg[object->head]->height);
				break;
			case GRAVITY_SWITCH: ML_bmp_or_cl(gravity_switch->tile[object->position], object->x-camera.x, object->y-camera.y, gravity_switch->width, gravity_switch->height); break;
			case CUBE: ML_bmp_or_cl(cube->tile[0], object->x-camera.x, object->y-camera.y, cube->width, cube->height); break;
			case CATERPILLAR: ML_bmp_or_cl(caterpillar[object->head][object->look]->tile[object->position], object->x-camera.x, object->y-camera.y, caterpillar[object->head][object->look]->width, caterpillar[object->head][object->look]->height); break;
			case SHOOTING_PLANT: ML_bmp_or_cl(plant[object->look]->tile[object->position], object->x-camera.x, object->y-camera.y, plant[object->look]->width, plant[object->look]->height); break;
			case BEE:
				ML_bmp_or_cl(bee[object->look]->tile[object->position], object->x-camera.x, object->y-camera.y, bee[object->look]->width, bee[object->look]->height);
				if(object->var < 6) ML_bmp_or_cl(bee_halo->tile[object->var/2+3], object->x-camera.x, object->y-camera.y, bee_halo->width, bee_halo->height);
				if(object->var >= TIME_BEE-2-3*3) ML_bmp_or_cl(bee_halo->tile[(object->var-(TIME_BEE-2-3*3))%3], object->x-camera.x, object->y-camera.y, bee_halo->width, bee_halo->height);
				break;
			case BULLET: ML_bmp_or_cl(bullet->tile[0], object->x-camera.x, object->y-camera.y, bullet->width, bullet->height); break;
			case HOMING_BULLET: ML_bmp_or_cl(bullet->tile[1], object->x-camera.x, object->y-camera.y, bullet->width, bullet->height); break;
			case PESTLE: ML_bmp_or_cl(pestle[object->head]->tile[object->position], object->x-camera.x, object->y-camera.y, pestle[object->head]->width, pestle[object->head]->height); break;
		}
	}
}
