#include <stdlib.h>
#include <fxlib.h>
#include "MonochromeLib.h"

#include "draw.h"
#include "engine.h"
#include "key.h"
#include "level.h"
#include "object.h"
#include "tileset.h"
#include "time.h"

/* public functions */
int engine_main(int id_level)
{
	int run=1;
	Camera camera = {0, 0};
	Level* level;
	if(draw_init()) return;
	while(run) {
		level = level_load(id_level);
		if(level) {
			switch(engine_loop(level)) {
				case QUIT: run = 0; break;
				case NEXT: id_level++; break;
			}
			level_free(level);
		} else run = 0;
	}
	draw_quit();
	return id_level;
}

void engine_intro(int page)
{
	int run=1, time=0, y=22;
	Tileset *statue, *duck, *chest, *congratulations;
	Level* level;
	Camera camera = {0, 10};

	if(draw_init()) return;
	statue = tileset_load(STATUE);
	duck = tileset_load(DUCK_UP_LEFT);
	chest = tileset_load(CHEST);
	level = level_load(0);
	while(key_down(K_EXE));
	while(run) {
		ML_clear_vram();
		draw_tileMap(level, camera);
		ML_bmp_or_cl(statue->tile[0], 5, 2, statue->width, statue->height);
		ML_bmp_or_cl(duck->tile[0], 110, 38, duck->width, duck->height);
		switch(page) {
			case 0:
				PrintMini(45, 10, "Hello my", MINI_OR);
				PrintMini(45, 17, "little minion.", MINI_OR);
				break;
			case 1:
				PrintMini(45, 10, "Bring me 40 golden", MINI_OR);
				PrintMini(45, 17, "eggs and I will", MINI_OR);
				PrintMini(45, 24, "make you rich.", MINI_OR);
				break;
			case 2:
				PrintMini(45, 10, "You now have", MINI_OR);
				PrintMini(45, 17, "the power to", MINI_OR);
				PrintMini(45, 24, "control gravity.", MINI_OR);
				break;
			case 3:
				PrintMini(45, 10, "Use the arrow keys", MINI_OR);
				PrintMini(45, 17, "to move.", MINI_OR);
				PrintMini(45, 24, "Pick up the egg to", MINI_OR);
				PrintMini(45, 31, "complete the level.", MINI_OR);
				break;
			case 4:
				PrintMini(45, 10, "Flip the gravity", MINI_OR);
				PrintMini(45, 17, "by pressing SHIFT.", MINI_OR);
				PrintMini(45, 24, "Press ALPHA+arrows", MINI_OR);
				PrintMini(45, 31, "to move the camera.", MINI_OR);
				break;
			case 6:
				PrintMini(45, 10, "Ah, good, I see", MINI_OR);
				PrintMini(45, 17, "you have returned", MINI_OR);
				PrintMini(45, 24, "with the eggs!", MINI_OR);
				break;
			case 7:
				if(time_getTicks() - time > 4) {
					if(y<38) y++;
					time = time_getTicks();
				}
				ML_bmp_or_cl(chest->tile[0], 67, y, chest->width, chest->height);
				PrintMini(45, 10, "Here.", MINI_OR);
				PrintMini(45, 17, "This is your reward.", MINI_OR);
				break;
			case 8:
				ML_bmp_or_cl(chest->tile[0], 67, 38, chest->width, chest->height);
				PrintMini(45, 10, "Now begone!", MINI_OR);
				PrintMini(45, 17, "I have an", MINI_OR);
				PrintMini(45, 24, "omelet to make!", MINI_OR);
				break;
		}
		ML_display_vram();
		if(key_down(K_EXIT)) {
			while(key_down(K_EXIT));
			run = 0;
		}
		if(key_down(K_EXE)) {
			while(key_down(K_EXE));
			page++;
		}
		if(page==5 || page == 9) run = 0;
	}
	level_free(level);
	tileset_free(duck);
	tileset_free(chest);
	tileset_free(statue);
	draw_quit();
	if(page == 9) {
		Confetti confetti[NB_CONFETTI];
		int i;
		char dx[] = {0, 1, 2, 3, 4, 1, 0, -1, -2, -3, -4, -1};
		for(i=0 ; i<NB_CONFETTI ; i++) {
			confetti[i].x = rand()%128;
			confetti[i].y = rand()%64-64;
			confetti[i].dy = rand()%2+1;
			confetti[i].position = rand()%12;
		}
		duck = tileset_load(HAT_DUCK);
		congratulations = tileset_load(CONGRATULATIONS);
		run = 1;
		while(run) {
			ML_clear_vram();
			for(i=0 ; i<NB_CONFETTI ; i++) ML_pixel(confetti[i].x, confetti[i].y, BLACK);
			ML_bmp_or_cl(congratulations->tile[0], 0, 3, congratulations->width, congratulations->height);
			ML_bmp_and_cl(duck->tile[1], (SCREEN_WIDTH-duck->width)/2, SCREEN_HEIGHT/2, duck->width, duck->height);
			ML_bmp_or_cl(duck->tile[0], (SCREEN_WIDTH-duck->width)/2, SCREEN_HEIGHT/2, duck->width, duck->height);
			ML_display_vram();
			if(time_getTicks() - time > 15) {
				for(i=0 ; i<NB_CONFETTI ; i++) {
					confetti[i].y = confetti[i].y+confetti[i].dy;
					if(confetti[i].y > 70) confetti[i].y = 0;
					confetti[i].x += dx[confetti[i].position];
					confetti[i].position = (confetti[i].position+1)%12;
				}
				time = time_getTicks();
			}
			if(key_down(K_EXE)) run = 2;
			else if(run == 2) run = 0;
		}
		tileset_free(congratulations);
		tileset_free(duck);
	}
}

/* private functions */
static Return_code engine_loop(Level* level)
{
	int i;
	Camera camera;
	Object* duck = object_get(level->object, DUCK);
	while(1) {
		if(key_down(K_EXIT)) {
			int key=0, cursor=0;
			Tileset *pause = tileset_load(PAUSE);
			if(!pause) return QUIT;
			while(key != KEY_CTRL_EXE) {
				ML_bmp_and_cl(pause->tile[0], (SCREEN_WIDTH-pause->width)/2, (SCREEN_HEIGHT-pause->height)/2, pause->width, pause->height);
				ML_bmp_or_cl(pause->tile[0], (SCREEN_WIDTH-pause->width)/2, (SCREEN_HEIGHT-pause->height)/2, pause->width, pause->height);
				ML_rectangle((SCREEN_WIDTH-pause->width)/2+5, (SCREEN_HEIGHT-pause->height)/2+17+cursor*13, (SCREEN_WIDTH-pause->width)/2+48, (SCREEN_HEIGHT-pause->height)/2+25+cursor*13, 1, BLACK, TRANSPARENT);
				GetKey(&key);
				if(key==KEY_CTRL_UP && cursor) cursor--;
				if(key==KEY_CTRL_DOWN && !cursor) cursor++;
			}
			tileset_free(pause);
			if(cursor) return QUIT;
		}

		engine_movements(level);

		if(!key_down(K_ALPHA)) {
			camera.x = duck->x - (SCREEN_WIDTH-TILE_WIDTH)/2;
			camera.y = duck->y - (SCREEN_HEIGHT-TILE_HEIGHT)/2;
		} else {
			if(key_down(K_UP)) camera.y--;
			if(key_down(K_DOWN)) camera.y++;
			if(key_down(K_LEFT)) camera.x--;
			if(key_down(K_RIGHT)) camera.x++;
		}
		if(camera.x < 0) camera.x = 0;
		if(camera.x > level->width*TILE_WIDTH-SCREEN_WIDTH) camera.x = level->width*TILE_WIDTH-SCREEN_WIDTH;
		if(camera.y < 0) camera.y = 0;
		if(camera.y > level->height*TILE_HEIGHT-SCREEN_HEIGHT) camera.y = level->height*TILE_HEIGHT-SCREEN_HEIGHT;

		ML_clear_vram();
		draw_tileMap(level, camera);
		draw_objects(level, camera);
		ML_display_vram();

		if(engine_collisions(level)) {
			for(i=500 ; i<1000 ; i++) {
				ML_filled_circle(duck->x+TILE_WIDTH/2-camera.x, duck->y+TILE_HEIGHT/2-camera.y, i/100+2, BLACK);
				ML_filled_circle(duck->x+TILE_WIDTH/2-camera.x, duck->y+TILE_HEIGHT/2-camera.y, i/100, WHITE);
				ML_display_vram();
			}
			return RETRY;
		}
		if(object_get(level->object, EGG)->position == 4) return NEXT;
	}
}

static void engine_movements(Level* level)
{
	static int time=0, i;
	Object* duck;
	if(time_getTicks() - time > 2) {
		engine_anim(level);
		engine_move_objects(level);
		duck = object_get(level->object, DUCK);
		if(!key_down(K_ALPHA)) {
			if(level->gravity == UP || level->gravity == DOWN) {
				if(key_down(K_LEFT)) { duck->look=LEFT; if(!engine_hitMap(level, duck, -1, 0)) duck->x--; }
				else if(key_down(K_RIGHT)) { duck->look=RIGHT; if(!engine_hitMap(level, duck, 1, 0)) duck->x++; }
				else duck->position = 0;
			} else {
				if(key_down(K_UP)) { duck->look=UP; if(!engine_hitMap(level, duck, 0, -1)) duck->y--; }
				else if(key_down(K_DOWN)) { duck->look=DOWN; if(!engine_hitMap(level, duck, 0, 1)) duck->y++; }
				else duck->position = 0;
			}
			if(engine_hitMap(level, duck, (level->gravity==LEFT?-1:level->gravity==RIGHT?1:0), (level->gravity==UP?-1:level->gravity==DOWN?1:0))) {
				if(key_down(K_SHIFT)) {
					duck->head = (duck->head+2)%4;
					level->gravity = (level->gravity+2)%4;
				}
			} else duck->position = 0;
		} else duck->position = 0;
		for(i=SPEED_FALLING ; i>0 ; i--) {
			switch(level->gravity) {
				case UP: if(!engine_hitMap(level, duck, 0, -i)) duck->y-=i, i=0; break;
				case DOWN: if(!engine_hitMap(level, duck, 0, i)) duck->y+=i, i=0; break;
				case LEFT: if(!engine_hitMap(level, duck, -i, 0)) duck->x-=i, i=0; break;
				case RIGHT: if(!engine_hitMap(level, duck, i, 0)) duck->x+=i, i=0; break;
			}
		}
		time = time_getTicks();
	}
}

static void engine_move_objects(Level* level)
{
	static int time=0;
	int i, j, n, min_distance, distance;
	char pestle_hitbox[] = {31, 27, 22, 18, 14, 10};
	char homing_bullet[28][2] = {{5, 0}, {5, 1}, {5, 2}, {4, 3}, {3, 4}, {2, 5}, {1, 5}, {0, 5}, {-1, 5}, {-2, 5}, {-3, 4}, {-4, 3}, {-5, 2}, {-5, 1}, {-5, 0}, {-5, -1}, {-5, -2}, {-4, -3}, {-3, -4}, {-2, -5}, {-1, -5}, {0, -5}, {1, -5}, {2, -5}, {3, -4}, {4, -3}, {5, -2}, {5, -1}};
	Object* object, *new, *duck = object_get(level->object, DUCK);
	if(time_getTicks() - time > 4) {
		engine_anim(level);
		for(i=0 ; level->object[i] ; i++) {
			object = level->object[i];
			switch(object->type) {
				case CUBE:
					if(engine_hitMap(level, object, object->dx, object->dy)) object->dx*=-1, object->dy*=-1;
					object->x+=object->dx;
					object->y+=object->dy;
					break;
				case CATERPILLAR:
					switch(object->head) {
						case UP: if(engine_hitMap(level, object, object->dx, 0) || !engine_hitMap(level, object, object->dx*TILE_WIDTH, TILE_HEIGHT)) object->dx*=-1, object->look = (object->look+2)%4; break;
						case DOWN: if(engine_hitMap(level, object, object->dx, 0) || !engine_hitMap(level, object, object->dx*TILE_WIDTH, -TILE_HEIGHT)) object->dx*=-1, object->look = (object->look+2)%4; break;
						case LEFT: if(engine_hitMap(level, object, 0, object->dy) || !engine_hitMap(level, object, TILE_WIDTH, object->dy*TILE_HEIGHT)) object->dy*=-1, object->look = (object->look+2)%4; break;
						case RIGHT: if(engine_hitMap(level, object, 0, object->dy) || !engine_hitMap(level, object, -TILE_WIDTH, object->dy*TILE_HEIGHT)) object->dy*=-1, object->look = (object->look+2)%4; break;
					}
					object->x+=object->dx;
					object->y+=object->dy;
					break;
				case SHOOTING_PLANT:
					if(object->var == TIME_PLANT-1) {
						object->var = 0;
						if(object->look == LEFT) {
							new = object_new(BULLET, object->x-1+SPEED_BULLET, object->y+5, UP, UP);
							new->dx = -SPEED_BULLET;
						} else {
							new = object_new(BULLET, object->x+13-SPEED_BULLET, object->y+5, UP, UP);
							new->dx = SPEED_BULLET;
						}
						level->object = object_add(level->object, new);
					}
					break;
				case BEE:
					if(duck->x < object->x) object->look = LEFT;
					else object->look = RIGHT;
					if(object->var == TIME_BEE-1) {
						object->var = 0;
						new = object_new(HOMING_BULLET, object->x+7, object->y+7, UP, UP);
						n = 0;
						min_distance = (duck->x-object->x-homing_bullet[0][0])*(duck->x-object->x-homing_bullet[0][0]) + (duck->y-object->y-homing_bullet[0][1])*(duck->y-object->y-homing_bullet[0][1]);
						for(j=0 ; j<28 ; j++) {
							distance = (duck->x-object->x-homing_bullet[j][0])*(duck->x-object->x-homing_bullet[j][0]) + (duck->y-object->y-homing_bullet[j][1])*(duck->y-object->y-homing_bullet[j][1]);
							if(distance < min_distance) {
								min_distance = distance;
								n = j;
							}
						}
						new->position = n;
						new->dx = homing_bullet[n][0];
						new->dy = homing_bullet[n][1];
						level->object = object_add(level->object, new);
					}
					break;
				case BULLET:
					object->x+=object->dx;
					if(object->x<0 || object->x+object->hitbox.x2>=level->width*TILE_WIDTH) {
						level->object = object_del(level->object, object);
						i--;
					} else if(engine_hitMap(level, object, 0, 0)) {
						level->object = object_del(level->object, object);
						i--;
					}
					break;
				case HOMING_BULLET:
					if(object->var == 6) {
						min_distance = (duck->x-object->x-homing_bullet[object->position][0])*(duck->x-object->x-homing_bullet[object->position][0]) + (duck->y-object->y-homing_bullet[object->position][1])*(duck->y-object->y-homing_bullet[object->position][1]);
						n = (object->position + 1)%28;
						distance = (duck->x-object->x-homing_bullet[n][0])*(duck->x-object->x-homing_bullet[n][0]) + (duck->y-object->y-homing_bullet[n][1])*(duck->y-object->y-homing_bullet[n][1]);
						if(distance < min_distance) object->position = n;
						n = (object->position + 27)%28;
						distance = (duck->x-object->x-homing_bullet[n][0])*(duck->x-object->x-homing_bullet[n][0]) + (duck->y-object->y-homing_bullet[n][1])*(duck->y-object->y-homing_bullet[n][1]);
						if(distance < min_distance) object->position = n;
						object->dx = homing_bullet[object->position][0];
						object->dy = homing_bullet[object->position][1];
						object->var = 0;
					} else object->var++;
					object->x+=object->dx;
					object->y+=object->dy;
					if(object->x<0 || object->y<0 || object->x+object->hitbox.x2>=level->width*TILE_WIDTH || object->y+object->hitbox.y2>=level->height*TILE_HEIGHT) {
						level->object = object_del(level->object, object);
						i--;
					} else if(engine_hitMap(level, object, 0, 0)) {
						level->object = object_del(level->object, object);
						i--;
					}
					break;
				case PESTLE:
					switch(object->head) {
						case UP: object->hitbox.y1 = object->hitbox.y2-pestle_hitbox[object->position]; break;
						case DOWN: object->hitbox.y2 = object->hitbox.y1+pestle_hitbox[object->position]; break;
						case LEFT: object->hitbox.x1 = object->hitbox.x2-pestle_hitbox[object->position]; break;
						case RIGHT: object->hitbox.x2 = object->hitbox.x1+pestle_hitbox[object->position]; break;
					}
					break;
			}
		}
		time = time_getTicks();
	}
}

static void engine_anim(Level* level)
{
	static int time=0;
	int i;
	Object* object;
	if(time_getTicks() - time > 8) {
		for(i=0 ; level->object[i] ; i++) {
			object = level->object[i];
			switch(object->type) {
				case DUCK: object->position++; if(object->position > 8) object->position = 1; break;
				case EGG: object->var = !object->var; if(object->position && object->var) object->position++; break;
				case GRAVITY_SWITCH: object->position = (object->position-object->var+8)%8; break;
				case CATERPILLAR: object->position++; if(object->position > 5) object->position = 0; break;
				case SHOOTING_PLANT: if(object->var<6 && !(object->var&1)) object->position = (object->position+1)%3; object->var = (object->var+1)%TIME_PLANT; break;
				case BEE: if(object->var&1) object->position = (object->position+1)%4; object->var = (object->var+1)%TIME_BEE; break;
				case PESTLE: object->var = (object->var+1)%TIME_PESTLE; if(object->var < 6) object->position = (object->position+1)%6; break;
			}
		}
		time = time_getTicks();
	}
}

static int engine_collisions(Level* level)
{
	int ret = 0;
	char item[4];
	int i, tmp, hit_GRAVITY_SWITCH=0;
	static int change_gravity=0;
	Object *object, *duck = object_get(level->object, DUCK);
	for(i=0 ; level->object[i] ; i++) {
		object = level->object[i];
		if(object_collision(duck, object)) {
			switch(object->type) {
				case GRAVITY_SWITCH:
					if(!change_gravity) {
						level->gravity = (level->gravity+object->var+4)%4;
						duck->head = (duck->head+object->var+4)%4;
						duck->look = (duck->look+object->var+4)%4;
						tmp=duck->hitbox.x1; duck->hitbox.x1=duck->hitbox.y1; duck->hitbox.y1=tmp; //rotate hitbox
						tmp=duck->hitbox.x2; duck->hitbox.x2=duck->hitbox.y2; duck->hitbox.y2=tmp; //rotate hitbox
						duck->x = object->x;
						duck->y = object->y;
						duck->y = object->y;
					}
					hit_GRAVITY_SWITCH=1;
					break;
				case EGG:
					if(duck->head == object->head && !object->position)
						object->position = 1;
					break;
				case CUBE:
				case CATERPILLAR:
				case BULLET:
				case HOMING_BULLET:
				case PESTLE:
					ret = RETRY;
					break;
			}
		}
	}
	change_gravity = hit_GRAVITY_SWITCH ? 1 : 0;
	
	/* if duck touch the map border */
	if(duck->x < 4 || duck->y < 4 || duck->x > (level->width-1)*TILE_WIDTH-4 || duck->y > (level->height-1)*TILE_HEIGHT-4)
		return RETRY;

	/* collisions with spikes */
	item[0] = level->map[((duck->y+duck->hitbox.y1)/TILE_HEIGHT)*level->width + (duck->x+duck->hitbox.x1)/TILE_WIDTH];
	item[1] = level->map[((duck->y+duck->hitbox.y2)/TILE_HEIGHT)*level->width + (duck->x+duck->hitbox.x1)/TILE_WIDTH];
	item[2] = level->map[((duck->y+duck->hitbox.y1)/TILE_HEIGHT)*level->width + (duck->x+duck->hitbox.x2)/TILE_WIDTH];
	item[3] = level->map[((duck->y+duck->hitbox.y2)/TILE_HEIGHT)*level->width + (duck->x+duck->hitbox.x2)/TILE_WIDTH];
	for(i=0 ; i<4 ; i++) {
		switch(item[i]) {
			case SPIKE_UP: if((duck->y+duck->hitbox.y2)%TILE_HEIGHT >= TILE_HEIGHT/2) return RETRY; break;
			case SPIKE_DOWN: if((duck->y+duck->hitbox.y1)%TILE_HEIGHT < TILE_HEIGHT/2) return RETRY; break;
			case SPIKE_LEFT: if((duck->x+duck->hitbox.x2)%TILE_WIDTH >= TILE_WIDTH/2) return RETRY; break;
			case SPIKE_RIGHT: if((duck->x+duck->hitbox.x1)%TILE_WIDTH < TILE_WIDTH/2) return RETRY; break;
		}
	}
	return ret;
}

static int engine_hitMap(Level* level, Object* object, int dx, int dy)
{
	char item;
	item = level->map[((object->y+object->hitbox.y1+dy)/TILE_HEIGHT)*level->width + (object->x+object->hitbox.x1+dx)/TILE_WIDTH];
	if(item != EMPTY && item != SPIKE_UP && item != SPIKE_DOWN && item != SPIKE_LEFT && item != SPIKE_RIGHT) return 1;
	item = level->map[((object->y+object->hitbox.y2+dy)/TILE_HEIGHT)*level->width + (object->x+object->hitbox.x1+dx)/TILE_WIDTH];
	if(item != EMPTY && item != SPIKE_UP && item != SPIKE_DOWN && item != SPIKE_LEFT && item != SPIKE_RIGHT) return 1;
	item = level->map[((object->y+object->hitbox.y1+dy)/TILE_HEIGHT)*level->width + (object->x+object->hitbox.x2+dx)/TILE_WIDTH];
	if(item != EMPTY && item != SPIKE_UP && item != SPIKE_DOWN && item != SPIKE_LEFT && item != SPIKE_RIGHT) return 1;
	item = level->map[((object->y+object->hitbox.y2+dy)/TILE_HEIGHT)*level->width + (object->x+object->hitbox.x2+dx)/TILE_WIDTH];
	if(item != EMPTY && item != SPIKE_UP && item != SPIKE_DOWN && item != SPIKE_LEFT && item != SPIKE_RIGHT) return 1;
	return 0;
}
