#include "MonochromeLib.h"

#include "draw.h"
#include "key.h"
#include "level.h"
#include "menu.h"
#include "tileset.h"
#include "time.h"

/* public functions */
void menu_title() 
{
	Tileset *title = tileset_load(TITLE);
	if(!title) return;
	while(key_down(K_EXE));
	while(!key_down(K_EXE)) {
		ML_clear_vram();
		menu_background();
		ML_bmp_and_cl(title->tile[0], (SCREEN_WIDTH-title->width)/2, (SCREEN_HEIGHT-title->height)/2, title->width, title->height);
		ML_bmp_or_cl(title->tile[0], (SCREEN_WIDTH-title->width)/2, (SCREEN_HEIGHT-title->height)/2, title->width, title->height);
		ML_display_vram();
	}
	tileset_free(title);
}

int menu_main(int cursor, int unlock)
{
	int i, x, y, time=0, ret=-1, first=1;
	Tileset *select_level, *button;

	select_level = tileset_load(SELECT_LEVEL);
	button = tileset_load(BUTTON);
	if(cursor > NB_LEVEL) cursor = NB_LEVEL;
	if(unlock > NB_LEVEL) unlock = NB_LEVEL;
	while(key_down(K_EXE));
	while(key_down(K_EXIT));
	while(ret<0) {
		if(time_getTicks() - time > 10) {
			if(key_down(K_EXIT)) ret = 0;
			if(key_down(K_EXE)) ret = cursor;
			if(key_down(K_LEFT) && cursor>1) cursor--;
			if(key_down(K_RIGHT) && cursor<unlock) cursor++;
			if(key_down(K_UP) && cursor>NB_ROW) cursor-=NB_ROW;
			if(key_down(K_DOWN) && cursor+NB_ROW<=unlock) cursor+=NB_ROW;
			if(cursor-first < NB_ROW  && first > 1) first -= NB_ROW;
			if(cursor-first >= NB_DISPLAYED_LEVEL-NB_ROW && first < NB_LEVEL-NB_DISPLAYED_LEVEL) first += NB_ROW;
			time = time_getTicks();
		}
		ML_clear_vram();
		menu_background();
		ML_bmp_and_cl(select_level->tile[0], (SCREEN_WIDTH-select_level->width)/2, 2, select_level->width, select_level->height);
		ML_bmp_or_cl(select_level->tile[0], (SCREEN_WIDTH-select_level->width)/2, 2, select_level->width, select_level->height);
		ML_rectangle(MENU_X-2, MENU_Y-2, MENU_X+MENU_WIDTH, MENU_Y+MENU_HEIGHT, 1, BLACK, WHITE);
		for(i=0 ; i<NB_DISPLAYED_LEVEL && i+first <= NB_LEVEL ; i++) {
			x = MENU_X+(i%NB_ROW)*(ITEM_WIDTH+1);
			y = MENU_Y+(i/NB_ROW)*(ITEM_HEIGHT+1);
			ML_rectangle(x, y, x+ITEM_WIDTH-1, y+ITEM_HEIGHT-1, 1, BLACK, WHITE);
			ML_bmp_or_cl(button->tile[i+first<=unlock?i+first:0], x+2, y+2, button->width, button->height);
		}
		x = MENU_X+((cursor-first)%NB_ROW)*(ITEM_WIDTH+1);
		y = MENU_Y+((cursor-first)/NB_ROW)*(ITEM_HEIGHT+1);
		if(cursor < first+NB_DISPLAYED_LEVEL) ML_rectangle(x, y, x+ITEM_WIDTH-1, y+ITEM_HEIGHT-1, 1, BLACK, XOR);
		ML_display_vram();
	}
	tileset_free(select_level);
	tileset_free(button);
	return ret;
}

/* private functions */
static void menu_background()
{
	static int time=0, x[NB_CIRCLE]={0, 33, 66, 99, 132}, y[NB_CIRCLE]={15, 42, 20, 40, 25}, size[NB_CIRCLE]={12, 20, 13, 15, 10};
	int i;
	if(time_getTicks() - time > 10) {
		for(i=0 ; i<NB_CIRCLE ; i++) {
			x[i]++;
			if(x[i]-20 > SCREEN_WIDTH)
				x[i] = -20;
		}
		time = time_getTicks();
	}
	for(i=0 ; i<NB_CIRCLE ; i++)
		ML_filled_circle(x[i], y[i], size[i] , BLACK);
}